# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from binascii import hexlify, unhexlify
from decimal import Decimal

from .sig import Sighash
from .script import (ScriptBuilder, P2wpkhV0Script, P2wshV0Script, P2shScript, NulldataScript, ScriptSig,
                     CoinBaseScriptSig, ScriptPubKey)
from ..lib.types import Immutable, Mutable, Jsonizable, HexSerializable, cached
from ..lib.parsing import Parser, TransactionParser, Stream
from btcpy.constants import BitcoinMainnet


# noinspection PyUnresolvedReferences
class Sequence(Immutable, HexSerializable):

    disable_flag_position = 31
    type_flag_position = 22
    MAX = 0xffffffff

    @staticmethod
    def max():
        return Sequence(Sequence.MAX)

    @staticmethod
    def create(seq, blocks=True, disable=False):
        if seq > 0xffff:
            raise ValueError('Sequence value too high: {}'.format(seq))
        flags = 0
        if not blocks:
            flags |= 1 << Sequence.type_flag_position
        if disable:
            flags |= 1 << Sequence.disable_flag_position
        return flags + seq

    def __init__(self, seq):
        object.__setattr__(self, 'seq', seq)

    @property
    def n(self):
        return self.seq & 0xffff

    def __str__(self):
        return str(self.seq)

    def __repr__(self):
        return 'Sequence({})'.format(self.seq)

    def is_active(self):
        return not (self.seq & (1 << Sequence.disable_flag_position))

    def is_time(self):
        return bool(self.seq & (1 << Sequence.type_flag_position))

    def is_blocks(self):
        return not self.is_time()

    def for_script(self):
        from .script import StackData
        return StackData.from_int(self.seq)

    @cached
    def serialize(self):
        return bytearray(self.seq.to_bytes(4, 'little'))


# noinspection PyUnresolvedReferences
class TxIn(Immutable, HexSerializable, Jsonizable):
    """
    :txid, the txid of the transaction being spent
    :txout, the output number of the output being spent
    :script_sig, a scriptSig
    :sequence, the sequence number of the TxIn
    """

    @classmethod
    def from_json(cls, dic):

        if 'coinbase' in dic:
            return CoinBaseTxIn(CoinBaseScriptSig(bytearray(unhexlify(dic['coinbase']['hex']))),
                                Sequence(int(dic['sequence'])))
        return cls(dic['txid'],
                   dic['vout'],
                   ScriptSig(bytearray(unhexlify(dic['scriptSig']['hex']))),
                   Sequence(int(dic['sequence'])))

    def __init__(self, txid: str, txout: int, script_sig: ScriptSig, sequence: Sequence, witness=None):
        object.__setattr__(self, 'txid', txid)
        object.__setattr__(self, 'txout', txout)
        object.__setattr__(self, 'script_sig', script_sig)
        object.__setattr__(self, 'sequence', sequence)
        object.__setattr__(self, 'witness', witness)

    def to_json(self):
        result = {'txid': self.txid,
                  'vout': self.txout,
                  'scriptSig': self.script_sig.to_json()}

        if self.witness is not None:
            result['txinwitness'] = self.witness.to_json()
        result['sequence'] = str(self.sequence)
        return result

    @cached
    def serialize(self):
        result = Stream()
        result << unhexlify(self.txid)[::-1]
        result << self.txout.to_bytes(4, 'little')
        result << Parser.to_varint(len(self.script_sig))
        result << self.script_sig
        result << self.sequence
        return result.serialize()

    def is_replaceable(self):
        return self.sequence.seq < (0xffffffff - 1)

    def is_standard(self, prev_script=None):

        if not self.script_sig.is_standard(isinstance(prev_script, P2shScript)):
            return False

        if prev_script is not None and not prev_script.is_standard():
            return False

        if isinstance(prev_script, P2shScript):
            sig_data = self.script_sig.get_data()
            if not sig_data:
                return False
            prev_script = ScriptBuilder.identify(sig_data[-1].data)

        if isinstance(prev_script, (P2wshV0Script, P2wpkhV0Script)):
            if self.witness is None:
                return False
            return self.witness.is_standard(p2wsh=isinstance(prev_script, P2wshV0Script))

        return True

    def __str__(self):
        return 'TxIn(txid={}, txout={}, script_sig={}, sequence={}, witness={})'.format(self.txid,
                                                                                        self.txout,
                                                                                        self.script_sig,
                                                                                        self.sequence,
                                                                                        self.witness)

    def to_mutable(self):
        return MutableTxIn(self.txid, self.txout, self.script_sig, self.sequence, self.witness)


# noinspection PyUnresolvedReferences
class MutableTxIn(Mutable, TxIn):

    def to_immutable(self):
        return TxIn(self.txid, self.txout, self.script_sig, self.sequence, self.witness)


# noinspection PyUnresolvedReferences
class CoinBaseTxIn(TxIn):

    sentinel_txid = '0' * 64
    sentinel_txout = 0xffffffff

    def __init__(self, script_sig, sequence, witness=None):
        super().__init__(CoinBaseTxIn.sentinel_txid, CoinBaseTxIn.sentinel_txout, script_sig, sequence, witness)

    def to_json(self):
        return {'coinbase': self.script_sig.to_json(),
                'sequence': str(self.sequence)}

    def is_standard(self, *args, **kwargs):
        return True


# noinspection PyUnresolvedReferences
class TxOut(Immutable, HexSerializable, Jsonizable):
    """
    : value - value of the output <int>
    : n - sequence number
    : script_pubkey - ScriptPubkey object
    """

    @classmethod
    def from_json(cls, dic, network=BitcoinMainnet):
        return cls(
            int(Decimal(dic['value']) * network.to_unit),
            dic['n'],
            ScriptBuilder.identify(bytearray(unhexlify(dic['scriptPubKey']['hex']))),
            network,
        )

    def __init__(self, value: int, n: int, script_pubkey: ScriptPubKey, network):
        object.__setattr__(self, 'value', value)
        object.__setattr__(self, 'n', n)
        object.__setattr__(self, 'script_pubkey', script_pubkey)
        object.__setattr__(self, 'network', network)

    @property
    def type(self):
        return self.script_pubkey.type

    @property
    def req_sigs(self):
        return self.script_pubkey.req_sigs

    def address(self):
        return self.script_pubkey.address()

    def is_standard(self):
        return self.script_pubkey.is_standard()

    def get_sig(self):
        pass

    def to_json(self):
        return {
            'value': str(Decimal(self.value) * self.network.from_unit),
            'n': self.n,
            'scriptPubKey': self.script_pubkey.to_json()
        }

    @cached
    def serialize(self):
        result = Stream()
        result << self.value.to_bytes(8, 'little')
        result << Parser.to_varint(len(self.script_pubkey))
        result << self.script_pubkey
        return result.serialize()

    def is_dust(self, size_to_relay_fee):
        return self.value < self.get_dust_threshold(size_to_relay_fee)

    def get_dust_threshold(self, size_to_relay_fee):

        if isinstance(self.script_pubkey, NulldataScript):
            return 0

        size = len(self.serialize())

        if isinstance(self.script_pubkey, (P2wpkhV0Script, P2wshV0Script)):
            # sum the sizes of the parts of a transaction input
            # with 75 % segwit discount applied to the script size.
            size += (32 + 4 + 1 + (107 // Witness.scale_factor) + 4)
        else:
            size += (32 + 4 + 1 + 107 + 4)

        return 3 * size_to_relay_fee(size)

    def __str__(self):
        return "TxOut(value={}, n={}, scriptPubKey='{}')".format(self.value, self.n, self.script_pubkey)


class PeercoinTxOut(TxOut):

    def get_dust_threshold(self, size_to_relay_fee):

        if isinstance(self.script_pubkey, NulldataScript):
            return 0

        return 0.01


# noinspection PyUnresolvedReferences
class Locktime(Immutable, HexSerializable):

    blocks_threshold = 500000000

    def __init__(self, n):
        object.__setattr__(self, 'n', n)

    def __str__(self):
        return 'Locktime({})'.format(self.n)

    def is_blocks(self):
        return 0 < self.n < Locktime.blocks_threshold

    def is_time(self):
        return self.n >= Locktime.blocks_threshold

    def is_active(self):
        return self.n > 0

    def to_json(self):
        pass

    def for_script(self):
        from .script import StackData
        return StackData.from_int(self.n)

    @cached
    def serialize(self):
        return bytearray(self.n.to_bytes(4, 'little'))


# noinspection PyUnresolvedReferences
class Witness(Immutable, HexSerializable, Jsonizable):

    scale_factor = 4
    max_redeem_script_size = 3600
    max_p2sh_push_ops = 100
    max_single_push_size = 80

    @classmethod
    def from_json(cls, string):
        pass

    def __init__(self, stack_items):
        object.__setattr__(self, 'items', stack_items)

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        yield from self.items

    def __add__(self, other):
        return Witness(self.items + other.items)

    def to_json(self):
        return [str(item) for item in self.items]

    @cached
    def serialize(self):
        result = Stream()
        result << Parser.to_varint(len(self.items))
        for item in self.items:
            result << item
        return result.serialize()

    def to_script_sig(self):
        return ScriptSig.from_stack_data(self.items)

    def is_standard(self, p2wsh=False):
        if p2wsh:
            if len(self.items[-1]) > Witness.max_redeem_script_size:
                return False
            if len(self.items) - 1 > Witness.max_p2sh_push_ops:
                return False
            for item in self.items[:-1]:
                if len(item) > Witness.max_single_push_size:
                    return False
        return True

    def __str__(self):
        return 'Witness([{}])'.format(', '.join('"{}"'.format(item) for item in self.items))


# noinspection PyUnresolvedReferences
class Transaction(Immutable, HexSerializable, Jsonizable):

    ''' assemble transaction
    : version - tranasction version (default 1)
    : inputs - list of tx inputs, as TxIn object
    : outputs - list of tx outputs as TxOut object
    : locktime - tx locktime, as Locktime object
    : txid - default None
    '''

    max_version = 2
    max_weight = 400000

    @classmethod
    def unhexlify(cls, string, tx_parser=TransactionParser, network=BitcoinMainnet):
        return cls.deserialize(bytearray(unhexlify(string)), tx_parser, network)

    @classmethod
    def deserialize(cls, string, tx_parser, network):
        parser = tx_parser(string, network)
        result = parser.get_next_tx(False)
        if parser:
            raise ValueError('Leftover data after transaction')
        return result

    @classmethod
    def from_json(cls, tx_json, network=BitcoinMainnet):
        return cls(
            version=tx_json['version'],
            locktime=Locktime(tx_json['locktime']),
            txid=tx_json['txid'],
            ins=[TxIn.from_json(txin_json) for txin_json in tx_json['vin']],
            outs=[TxOut.from_json(txout_json) for txout_json in tx_json['vout']],
            network=network,
        )

    def __init__(self, version: int, ins: list, outs: list,
                 locktime: Locktime, network, txid: str=None) -> None:

        object.__setattr__(self, 'version', version)
        object.__setattr__(self, 'ins', tuple(ins))
        object.__setattr__(self, 'outs', tuple(outs))
        object.__setattr__(self, 'locktime', locktime)
        object.__setattr__(self, '_txid', txid)
        object.__setattr__(self, 'network', network)

        if txid != self.txid and txid is not None:
            raise ValueError('txid {} does not match transaction data {}'.format(txid, self.hexlify()))

    def _base_size(self):
        return len(self.serialize())

    @cached
    def hash(self):
        stream = Stream()
        stream << self
        return hexlify(stream.hash256()[::-1]).decode()

    @property
    def txid(self):
        return self.hash()

    @property
    @cached
    def normalized_id(self):
        mutable = self.to_mutable()
        for txin in mutable.ins:
            txin.script_sig = ScriptSig.empty()
        return mutable.txid

    @property
    def size(self):
        return len(self.serialize())

    @property
    def weight(self):
        return self._base_size() * 3 + self.size

    @property
    def vsize(self):
        from math import ceil
        return ceil(self.weight / 4)

    def to_json(self):
        return {
            'hex': self.hexlify(),
            'txid': self.txid,
            'hash': self.hash(),
            'size': self.size,
            'vsize': self.vsize,
            'version': self.version,
            'locktime': self.locktime.n,
            'vin': [txin.to_json() for txin in self.ins],
            'vout': [txout.to_json() for txout in self.outs],
        }

    @cached
    def serialize(self):
        from itertools import chain
        result = Stream()
        result << self.version.to_bytes(4, 'little')
        result << Parser.to_varint(len(self.ins))
        # the most efficient way to flatten a list in python
        result << bytearray(chain.from_iterable(txin.serialize() for txin in self.ins))
        result << Parser.to_varint(len(self.outs))
        # the most efficient way to flatten a list in python
        result << bytearray(chain.from_iterable(txout.serialize() for txout in self.outs))
        result << self.locktime
        return result.serialize()

    def is_replaceable(self):
        return any(txin.is_replaceable() for txin in self.ins)

    def is_standard(self, size_to_relay_fee, prev_scripts=None):

        if len(prev_scripts) != len(self.ins):
            raise ValueError('Prev scripts provided are a different number than txins')

        if not 1 <= self.version <= self.max_version:
            return False

        if self.weight > self.max_weight:
            return False

        for prev, txin in zip(prev_scripts, self.ins):
            if not txin.is_standard(prev):
                return False

        nulldata = 0
        for out in self.outs:
            if not out.script_pubkey.is_standard():
                return False
            if isinstance(out.script_pubkey, NulldataScript):
                nulldata += 1
                if nulldata > 1:
                    return False
                if out.is_dust(size_to_relay_fee):
                    return False

        return True

    def is_coinbase(self):
        return len(self.ins) == 1 and isinstance(self.ins[0], CoinBaseTxIn)

    def to_mutable(self):
        return MutableTransaction(
            self.version,
            [txin.to_mutable() for txin in self.ins],
            self.outs,
            self.locktime,
            self.network,
        )

    def get_digest_preimage(self, index, prev_script, sighash=Sighash('ALL')):

        # TODO: manage codeseparator

        # print([str(inp) for inp in self.ins])
        # print([str(out) for out in self.outs])
        # print('Computing digest for input {}...'.format(index))
        throwaway = self.to_mutable()
        for i in range(len(throwaway.ins)):
            # we are signing input i so we create
            # empty scriptsig for every txin unless i == j
            throwaway.ins[i].script_sig = ScriptSig.empty()
            if i == index:
                throwaway.ins[i].script_sig = prev_script

        if sighash in ('NONE', 'SINGLE'):

            if sighash == 'NONE':
                throwaway.outs = []

            elif sighash == 'SINGLE':
                if index >= len(throwaway.outs):
                    raise ValueError('TxIn index greater than number of outputs and SIGHASH_SINGLE was chosen!')
                matching_out = throwaway.outs[index]
                throwaway.outs = [TxOut(0xffffffffffffffff, i, ScriptPubKey.empty()) for i in range(index)]
                throwaway.outs.append(matching_out)

            # so that others can replace
            for i in range(len(throwaway.ins)):
                if i != index:
                    throwaway.ins[i].sequence = Sequence(0)

        if sighash.anyone:
            # remove all other inputs completely
            throwaway.ins = [throwaway.ins[index]]

        to_hash = Stream()
        to_hash << throwaway
        to_hash << sighash

        # print('SIGHASH: {}'.format(spend.sighash))
        # print(self)

        return to_hash

    def get_digest(self, txin, prev_script, sighash=Sighash('ALL')):
        return self.get_digest_preimage(txin, prev_script, sighash).hash256()

    def __str__(self):
        return ('Transaction(version={}, '
                'ins=[{}], '
                'outs=[{}], '
                'locktime={}, '.format(self.version,
                                       ', '.join(str(txin) for txin in self.ins),
                                       ', '.join(str(out) for out in self.outs),
                                       self.locktime,))

    def __eq__(self, other):
        return self.hash() == other.hash()


class PeercoinTx(Transaction):

    def __init__(self, version: int, timestamp: int, ins: list, outs: list,
                 locktime: Locktime, network, txid: str=None) -> None:

        object.__setattr__(self, 'version', version)
        object.__setattr__(self, 'timestamp', timestamp)
        object.__setattr__(self, 'ins', tuple(ins))
        object.__setattr__(self, 'outs', tuple(outs))
        object.__setattr__(self, 'locktime', locktime)
        object.__setattr__(self, '_txid', txid)
        object.__setattr__(self, 'network', network)

        if txid != self.txid and txid is not None:
            raise ValueError('txid {} does not match transaction data {}'.format(txid, self.hexlify()))

    @classmethod
    def from_json(cls, tx_json, network):
        return cls(
            version=tx_json['version'],
            timestamp=tx_json['time'],
            locktime=Locktime(tx_json['locktime']),
            txid=tx_json['txid'],
            ins=[TxIn.from_json(txin_json) for txin_json in tx_json['vin']],
            outs=[TxOut.from_json(txout_json) for txout_json in tx_json['vout']],
            network=network,
        )

    def to_json(self):
        return {
            'hex': self.hexlify(),
            'txid': self.txid,
            'hash': self.hash(),
            'size': self.size,
            'vsize': self.vsize,
            'version': self.version,
            'timestamp': self.timestamp,
            'locktime': self.locktime.n,
            'vin': [txin.to_json() for txin in self.ins],
            'vout': [txout.to_json() for txout in self.outs],
        }

    @cached
    def serialize(self):
        from itertools import chain
        result = Stream()
        result << self.version.to_bytes(4, 'little')
        result << self.timestamp.to_bytes(4, 'little')
        result << Parser.to_varint(len(self.ins))
        # the most efficient way to flatten a list in python
        result << bytearray(chain.from_iterable(txin.serialize() for txin in self.ins))
        result << Parser.to_varint(len(self.outs))
        # the most efficient way to flatten a list in python
        result << bytearray(chain.from_iterable(txout.serialize() for txout in self.outs))
        result << self.locktime
        return result.serialize()

    def to_mutable(self):
        return PeercoinMutableTx(
            version=self.version,
            timestamp=self.timestamp,
            ins=[txin.to_mutable() for txin in self.ins],
            outs=self.outs,
            locktime=self.locktime,
            network=self.network,
        )

    def __str__(self):
        return ('PeercoinTx(version={}, '
                'ins=[{}], '
                'outs=[{}], '
                'locktime={}, '
                'timestamp={} '.format(self.version,
                                       ', '.join(str(txin) for txin in self.ins),
                                       ', '.join(str(out) for out in self.outs),
                                       self.locktime,
                                       self.timestamp))


class MutableTransaction(Mutable, Transaction):

    def __init__(self, version: int, ins: list,
                 outs: list, locktime: Locktime, network) -> None:

        super().__init__(version, ins, outs, locktime, network)
        ins = []
        for txin in self.ins:
            if isinstance(txin, MutableTxIn):
                ins.append(txin)
            elif isinstance(txin, TxIn):
                ins.append(txin.to_mutable())
            else:
                raise ValueError('Expected objects of type `TxIn` or `MutableTxIn`, got {} instead'.format(type(txin)))
        self.ins = ins
        self.outs = list(self.outs)

    def to_immutable(self):
        return Transaction(
            self.version,
            [txin.to_immutable() for txin in self.ins],
            self.outs,
            self.locktime,
            self.constants,
        )

    def to_segwit(self):
        return MutableSegWitTransaction(self.version, self.ins, self.outs, self.locktime)

    @classmethod
    def deserialize(cls, string, tx_parser, constants):
        parser = tx_parser(string, constants)
        result = parser.get_next_tx(True)
        if parser:
            raise ValueError('Leftover data after transaction')
        return result

    def spend_single(self, index, txout, solver):

        sighashes = solver.get_sighashes()
        prev_script = solver.get_prev_script() if solver.has_prev_script() else txout.script_pubkey

        if len(sighashes) == 0:
            script_sig, witness = solver.solve()
        elif len(sighashes) == 1:
            script_sig, witness = solver.solve(self.get_digest(index, prev_script, sighashes[0]))
        else:
            digests = []
            for sighash in sighashes:
                digests.append(self.get_digest(index, prev_script, sighash))
            script_sig, witness = solver.solve(*digests)

        if witness:
            raise ValueError('Trying to spend segwit output with non-segwit transaction!')

        self.ins[index].script_sig = script_sig

    def spend(self, txouts, solvers):
        if any(solver.solves_segwit() for solver in solvers):
            return self.to_segwit().spend(txouts, solvers)
        if len(solvers) != len(self.ins) or len(txouts) != len(solvers):
            raise ValueError('{} solvers and {} txouts provided for {} inputs'.format(len(solvers),
                                                                                      len(txouts),
                                                                                      len(self.ins)))
        for i, (txout, solver) in enumerate(zip(txouts, solvers)):
            self.spend_single(i, txout, solver)

        return self.to_immutable()


class PeercoinMutableTx(PeercoinTx, MutableTransaction):

    def __init__(self, version: int, timestamp: int, ins: list,
                 outs: list, locktime: Locktime, network) -> None:

        super().__init__(version, timestamp, ins, outs, locktime, network)
        ins = []
        for txin in self.ins:
            if isinstance(txin, MutableTxIn):
                ins.append(txin)
            elif isinstance(txin, TxIn):
                ins.append(txin.to_mutable())
            else:
                raise ValueError('Expected objects of type `TxIn` or `MutableTxIn`, got {} instead'.format(type(txin)))
        self.ins = ins
        self.outs = list(self.outs)

    def to_immutable(self):
        return PeercoinTx(
            self.version,
            self.timestamp,
            [txin.to_immutable() for txin in self.ins],
            self.outs,
            self.locktime,
            self.network,
        )

    def to_segwit(self):
        raise NotImplementedError("Peercoin doesn't have SegWit.")


class SegWitTransaction(Immutable, HexSerializable, Jsonizable):

    marker = 0x00
    flag = 0x01
    byte_marker = bytearray([marker])
    byte_flag = bytearray([flag])

    @staticmethod
    def unhexlify(string):
        tx = Transaction.unhexlify(string)
        return SegWitTransaction(tx.version, tx.ins, tx.outs, tx.locktime)

    @staticmethod
    def from_json(string):
        tx = Transaction.from_json(string)
        return SegWitTransaction(tx.version, tx.ins, tx.outs, tx.locktime)

    def __init__(self, version, ins, outs, locktime, txid=None):
        object.__setattr__(self, 'transaction', Transaction(version, ins, outs, locktime, txid))

    def __getattr__(self, item):
        return getattr(self.transaction, item)

    @cached
    def serialize(self):
        from itertools import chain
        classic = self.transaction.serialize()
        version, middle, locktime = classic[:4], classic[4:-4], classic[-4:]
        witness = bytearray(chain.from_iterable(txin.witness.serialize() for txin in self.ins))
        return version + self.__class__.byte_marker + self.__class__.byte_flag + middle + witness + locktime

    @property
    def normalized_id(self):
        return self.txid

    @cached
    def hash(self):
        stream = Stream()
        stream << self
        return hexlify(stream.hash256()[::-1]).decode()

    def hexlify(self):
        return super().hexlify()

    def to_mutable(self):
        return MutableSegWitTransaction(self.version,
                                        [txin.to_mutable() for txin in self.ins],
                                        self.outs,
                                        self.locktime)

    def get_digest(self, index, prev_script, sighash=Sighash('ALL')):
        return self.transaction.get_digest(index, prev_script, sighash)

    def get_segwit_digest(self, index, prev_script, prev_amount, sighash=Sighash('ALL')):
        return self._get_segwit_digest_preimage(index, prev_script, prev_amount, sighash).hash256()

    def _get_segwit_digest_preimage(self, index, prev_script, prev_amount, sighash=Sighash('ALL')):

        # TODO reinsert partial caching to avoid quadratic hashing

        hash_prevouts = bytearray([0] * 32)
        hash_sequence = bytearray([0] * 32)
        hash_outputs = bytearray([0] * 32)

        if not sighash.anyone:
            # if cache['hash_prevouts'] is None:
            # cache['hash_prevouts'] = self._hash_prevouts()
            # hash_prevouts = cache['hash_prevouts']
            hash_prevouts = self._hash_prevouts()

        if not sighash.anyone and (sighash not in ('SINGLE', 'NONE')):
            # if cache['hash_sequence'] is None:
            #     cache['hash_sequence'] = self._hash_sequence()
            # hash_sequence = cache['hash_sequence']
            hash_sequence = self._hash_sequence()

        if sighash not in ('SINGLE', 'NONE'):
            # if cache['hash_outputs'] is None:
            #     cache['hash_outputs'] = self._hash_outputs()
            # hash_outputs = cache['hash_outputs']
            hash_outputs = self._hash_outputs()
        elif sighash == 'SINGLE' and index < len(self.outs):
            hash_outputs = (Stream() << self.outs[index]).hash256()

        if isinstance(prev_script, P2wpkhV0Script):
            script_code = prev_script.get_scriptcode()
        else:
            script_code = prev_script.to_stack_data()

        curr_in = self.ins[index]

        result = Stream()

        result << self.version.to_bytes(4, 'little')
        result << hash_prevouts
        result << hash_sequence
        result << unhexlify(curr_in.txid)[::-1]
        result << curr_in.txout.to_bytes(4, 'little')
        result << script_code
        result << prev_amount.to_bytes(8, 'little')
        result << curr_in.sequence
        result << hash_outputs
        result << self.locktime
        result << sighash

        # print('Version: {}'.format(hexlify(self.version.to_bytes(4, 'little'))))
        # print('Hash prevouts: {}'.format(hexlify(hash_prevouts)))
        # print('Hash sequence: {}'.format(hexlify(hash_sequence)))
        # print('Outpoint txid: {}'.format(hexlify(unhexlify(curr_in.txid)[::-1])))
        # print('Outpoint vout: {}'.format(hexlify(curr_in.txout.to_bytes(4, 'little'))))
        # print('Script code: {}'.format(script_code.hexlify()))
        # print('Prev amount: {}'.format(hexlify(spend.prev_amount.to_bytes(8, 'little'))))
        # print('Sequence: {}'.format(curr_in.sequence.hexlify()))
        # print('Hash outputs: {}'.format(hexlify(hash_outputs)))
        # print('Locktime: {}'.format(self.locktime.hexlify()))
        # print('Sighash: {}'.format(spend.sighash.hexlify()))
        # print('Digest: {}'.format(hexlify(result.hash())))

        return result

    def _hash_prevouts(self):
        result = Stream()
        for txin in self.ins:
            result << unhexlify(txin.txid)[::-1]
            result << txin.txout.to_bytes(4, 'little')
        return result.hash256()

    def _hash_sequence(self):
        result = Stream()
        for txin in self.ins:
            result << txin.sequence
        return result.hash256()

    def _hash_outputs(self):
        result = Stream()
        for out in self.outs:
            result << out
        return result.hash256()

    # the following properties must be redefined,
    # it does not work if one calls them on the inner transaction
    # the only difference with what is defined in Transaction
    # is that here _base_size() extracts the inner transaction's base_size
    @property
    def size(self):
        return len(self.serialize())

    @property
    def weight(self):
        return self._base_size() * 3 + self.size

    @property
    def vsize(self):
        from math import ceil
        return ceil(self.weight / 4)

    def to_json(self):
        return {'hex': self.hexlify(),
                'txid': self.txid,
                'hash': self.hash(),
                'size': self.size,
                'vsize': self.vsize,
                'version': self.version,
                'locktime': self.locktime.n,
                'vin': [txin.to_json() for txin in self.ins],
                'vout': [txout.to_json() for txout in self.outs]}

    def __str__(self):
        return ('SegWitTransaction(version={}, '
                'ins=[{}], '
                'outs=[{}], '
                'locktime={})'.format(self.version,
                                      ', '.join(str(txin) for txin in self.ins),
                                      ', '.join(str(out) for out in self.outs),
                                      self.locktime))

    def __eq__(self, other):
        return self.hash() == other.hash()


class MutableSegWitTransaction(Mutable, SegWitTransaction):

    def __init__(self, version, ins, outs, locktime, txid=None):
        super().__init__(version, ins, outs, locktime, txid)
        ins = []
        for txin in self.ins:
            if isinstance(txin, MutableTxIn):
                ins.append(ins)
            elif isinstance(txin, TxIn):
                ins.append(txin.to_mutable())
            else:
                raise ValueError('Expected objects of type `TxIn` or `MutableTxIn`, got {} instead'.format(type(txin)))
        self.transaction = self.transaction.to_mutable()

    def to_immutable(self):
        return SegWitTransaction(self.version,
                                 [txin.to_immutable() for txin in self.ins],
                                 self.outs,
                                 self.locktime)

    def spend_single(self, index, txout, solver):

        sighashes = solver.get_sighashes()
        prev_script = solver.get_prev_script() if solver.has_prev_script() else txout.script_pubkey

        if len(sighashes) == 0:
            script_sig, witness = solver.solve()
        elif len(sighashes) == 1:
            if solver.solves_segwit():
                digest = self.get_segwit_digest(index, prev_script, txout.value, sighashes[0])
                script_sig, witness = solver.solve(digest)
            else:
                digest = self.get_digest(index, prev_script, sighashes[0])
                script_sig, witness = solver.solve(digest)

        else:
            digests = []
            if solver.solves_segwit():
                for sighash in sighashes:
                    digests.append(self.get_segwit_digest(index,
                                                          prev_script,
                                                          txout.value,
                                                          sighash))
            else:
                for sighash in sighashes:
                    digests.append(self.get_digest(index, prev_script, sighash))
            script_sig, witness = solver.solve(*digests)

        self.ins[index].script_sig = script_sig
        self.ins[index].witness = witness

    def spend(self, txouts, solvers):

        if len(solvers) != len(self.ins) or len(txouts) != len(solvers):
            raise ValueError('{} solvers, {} txouts provided for {} inputs'.format(len(solvers),
                                                                                   len(txouts),
                                                                                   len(self.ins)))

        for i, (txout, solver) in enumerate(zip(txouts, solvers)):
            self.spend_single(i, txout, solver)

        return self.to_immutable()




# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from abc import ABCMeta, abstractmethod


class WrongScriptType(Exception):
    pass


class BaseAddress(metaclass=ABCMeta):

    @staticmethod
    def is_valid(network, string):
        from ..lib.codecs import CouldNotDecode

        try:
            Address.from_string(network, string)
            return True
        except CouldNotDecode:
            try:
                SegWitAddress.from_string(network, string)
                return True
            except CouldNotDecode:
                return False

    @staticmethod
    @abstractmethod
    def get_codec():
        raise NotImplemented

    @classmethod
    @abstractmethod
    def from_script(cls, network, script):
        raise NotImplemented

    @classmethod
    @abstractmethod
    def get_type(cls):
        raise NotImplemented

    @classmethod
    def from_string(cls, network, string):
        return cls.get_codec().decode(network, string)

    @classmethod
    def hash_length(cls):
        raise NotImplemented

    def __init__(self, network, hashed_data):

        if len(hashed_data) != self.__class__.hash_length():
            raise ValueError('Hashed data must be {}-bytes long, length: {}'.format(self.__class__.hash_length(), len(hashed_data)))

        self.network = network
        self.hash = hashed_data

    def __str__(self):
        return self.__class__.get_codec().encode(self)


class Address(BaseAddress, metaclass=ABCMeta):

    @staticmethod
    def get_codec():
        from ..lib.codecs import Base58Codec
        return Base58Codec

    def __eq__(self, other):
        return (self.network, self.hash) == (other.network, other.hash)


class SegWitAddress(BaseAddress, metaclass=ABCMeta):

    @staticmethod
    def get_codec():
        from ..lib.codecs import Bech32Codec
        return Bech32Codec

    def __init__(self, network, hashed_data, version):
        super().__init__(network, hashed_data)
        self.version = version

    def __eq__(self, other):
        return (self.network, self.hash, self.version) == (other.network, other.hash, other.version)


class P2pkhAddress(Address):

    @classmethod
    def get_type(cls):
        return 'p2pkh'

    @classmethod
    def from_script(cls, network, script):
        from .script import P2pkhScript
        # can't use isinstance here: P2wpkhScript is child of P2pkhScript
        if script.__class__ is not P2pkhScript:
            raise WrongScriptType('Trying to produce P2pkhAddress from {} script'.format(script.__class__.__name__))

        return cls(network, script.pubkeyhash)

    @classmethod
    def hash_length(cls):
        return 20


class P2shAddress(Address):

    @classmethod
    def get_type(cls):
        return 'p2sh'

    @classmethod
    def from_script(cls, network, script):
        from .script import P2shScript
        # can't use isinstance here: P2wshScript is child of P2shScript
        if script.__class__ is P2shScript:
            return cls(network, script.scripthash)
        return cls(network, script.p2sh_hash())

    @classmethod
    def hash_length(cls):
        return 20


class P2wpkhAddress(SegWitAddress):

    @classmethod
    def get_type(cls):
        return 'p2wpkh'

    @classmethod
    def from_script(cls, network, script):
        from .script import P2wpkhScript
        if not isinstance(script, P2wpkhScript):
            raise WrongScriptType('Trying to produce P2pkhAddress from {} script'.format(script.__class__.__name__))

        return cls(network, script.pubkeyhash, script.__class__.get_version())

    @classmethod
    def hash_length(cls):
        return 20


class P2wshAddress(SegWitAddress):

    @classmethod
    def get_type(cls):
        return 'p2wsh'

    @classmethod
    def from_script(cls, network, script):
        from .script import P2wshScript
        version = script.__class__.get_version()
        if isinstance(script, P2wshScript):
            hashed_data = script.scripthash
        else:
            hashed_data = script.p2wsh_hash()
        return cls(network, hashed_data, version)

    @classmethod
    def hash_length(cls):
        return 32



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.


from binascii import unhexlify, hexlify

from ..lib.parsing import BlockParser, BlockHeaderParser, Parser, TransactionParser
from ..lib.types import Immutable, Jsonizable, HexSerializable, cached
from btcpy.constants import BitcoinMainnet


# noinspection PyUnresolvedReferences
class Block(Immutable, Jsonizable, HexSerializable):

    max_difficulty = 0xffff0000000000000000000000000000000000000000000000000000

    def __init__(self, header, txns):
        object.__setattr__(self, 'header', header)
        object.__setattr__(self, 'txns', txns)

    @staticmethod
    def unhexlify(string, tx_parser=TransactionParser, network=BitcoinMainnet):
        return Block.deserialize(bytearray(unhexlify(string)), tx_parser, network)

    @staticmethod
    def deserialize(string, tx_parser, network):
        parser = BlockParser(string, tx_parser, network)
        header = parser.get_block_header()
        txns = parser.get_txns()
        return Block(header, txns)

    @cached
    def serialize(self):
        txns_arr = bytearray()
        for tx in self.txns:
            txns_arr += tx.serialize()
        return self.header.serialize() + Parser.to_varint(len(self.txns)) + txns_arr

    @cached
    def hash(self):
        from hashlib import sha256
        return hexlify(sha256(sha256(self.header.serialize()).digest()).digest()[::-1]).decode()

    @classmethod
    def from_json(cls, string):
        pass

    def to_json(self):
        return {
            'header': self.header.to_json(),
            'difficulty': self.bits_to_diff(),
            'hash': self.hash(),
            'txn_count': len(self.txns),
            'txns': {tx.txid for tx in self.txns}}

    def bits_to_diff(self):
        bits = self.header.bits
        n_size = int(bits, 16) >> 24
        current_target = int(bits, 16) & 0x007fffff
        if n_size <= 3:
            current_target >>= (8 * (3 - n_size))

        else:
            current_target <<= (8 * (n_size - 3))

        return Block.max_difficulty / current_target

    def __eq__(self, other):
        return self.hash() == other.hash()


# noinspection PyUnresolvedReferences
class BlockHeader(Immutable, Jsonizable, HexSerializable):

    def __init__(self, version, prev_block, merkle_root, timestamp, bits, nonce):
        object.__setattr__(self, 'version', version)
        object.__setattr__(self, 'prev_block', prev_block)
        object.__setattr__(self, 'merkle_root', merkle_root)
        object.__setattr__(self, 'timestamp', timestamp)
        object.__setattr__(self, 'bits', bits)
        object.__setattr__(self, 'nonce', nonce)

    @staticmethod
    def unhexlify(string):
        return BlockHeader.deserialize(bytearray(unhexlify(string)))

    @staticmethod
    def deserialize(string):
        parser = BlockHeaderParser(string)
        result = parser.get_block_header()
        return result

    @cached
    def serialize(self):
        version = bytearray(self.version.to_bytes(4, 'little'))
        prev_block = bytearray(unhexlify(self.prev_block)[::-1])
        merkle_root = bytearray(unhexlify(self.merkle_root)[::-1])
        timestamp = bytearray((self.timestamp.to_bytes(4, 'little')))
        bits = bytearray(unhexlify(self.bits)[::-1])
        nonce = bytearray(self.nonce.to_bytes(4, 'little'))
        return version + prev_block + merkle_root + timestamp + bits + nonce

    @classmethod
    def from_json(cls, string):
        pass

    def to_json(self):
        return {
            'version': self.version,
            'prev_block': self.prev_block,
            'merkle_root': self.merkle_root,
            'timestamp': self.timestamp,
            'bits': self.bits,
            'nonce': self.nonce}



class OpCodeConverter(object):

    opcodes = [('OP_0', 0),
                   ('OP_PUSHDATA1', 76),
                   ('OP_PUSHDATA2', 77),
                   ('OP_PUSHDATA4', 78),
                   ('OP_1NEGATE', 79),
                   ('OP_RESERVED', 80),
                   ('OP_1', 81),
                   ('OP_2', 82),
                   ('OP_3', 83),
                   ('OP_4', 84),
                   ('OP_5', 85),
                   ('OP_6', 86),
                   ('OP_7', 87),
                   ('OP_8', 88),
                   ('OP_9', 89),
                   ('OP_10', 90),
                   ('OP_11', 91),
                   ('OP_12', 92),
                   ('OP_13', 93),
                   ('OP_14', 94),
                   ('OP_15', 95),
                   ('OP_16', 96),
                   ('OP_NOP', 97),
                   ('OP_VER', 98),
                   ('OP_IF', 99),
                   ('OP_NOTIF', 100),
                   ('OP_VERIF', 101),
                   ('OP_VERNOTIF', 102),
                   ('OP_ELSE', 103),
                   ('OP_ENDIF', 104),
                   ('OP_VERIFY', 105),
                   ('OP_RETURN', 106),
                   ('OP_TOALTSTACK', 107),
                   ('OP_FROMALTSTACK', 108),
                   ('OP_2DROP', 109),
                   ('OP_2DUP', 110),
                   ('OP_3DUP', 111),
                   ('OP_2OVER', 112),
                   ('OP_2ROT', 113),
                   ('OP_2SWAP', 114),
                   ('OP_IFDUP', 115),
                   ('OP_DEPTH', 116),
                   ('OP_DROP', 117),
                   ('OP_DUP', 118),
                   ('OP_NIP', 119),
                   ('OP_OVER', 120),
                   ('OP_PICK', 121),
                   ('OP_ROLL', 122),
                   ('OP_ROT', 123),
                   ('OP_SWAP', 124),
                   ('OP_TUCK', 125),
                   ('OP_CAT', 126),
                   ('OP_SUBSTR', 127),
                   ('OP_LEFT', 128),
                   ('OP_RIGHT', 129),
                   ('OP_SIZE', 130),
                   ('OP_INVERT', 131),
                   ('OP_AND', 132),
                   ('OP_OR', 133),
                   ('OP_XOR', 134),
                   ('OP_EQUAL', 135),
                   ('OP_EQUALVERIFY', 136),
                   ('OP_RESERVED1', 137),
                   ('OP_RESERVED2', 138),
                   ('OP_1ADD', 139),
                   ('OP_1SUB', 140),
                   ('OP_2MUL', 141),
                   ('OP_2DIV', 142),
                   ('OP_NEGATE', 143),
                   ('OP_ABS', 144),
                   ('OP_NOT', 145),
                   ('OP_0NOTEQUAL', 146),
                   ('OP_ADD', 147),
                   ('OP_SUB', 148),
                   ('OP_MUL', 149),
                   ('OP_DIV', 150),
                   ('OP_MOD', 151),
                   ('OP_LSHIFT', 152),
                   ('OP_RSHIFT', 153),
                   ('OP_BOOLAND', 154),
                   ('OP_BOOLOR', 155),
                   ('OP_NUMEQUAL', 156),
                   ('OP_NUMEQUALVERIFY', 157),
                   ('OP_NUMNOTEQUAL', 158),
                   ('OP_LESSTHAN', 159),
                   ('OP_GREATERTHAN', 160),
                   ('OP_LESSTHANOREQUAL', 161),
                   ('OP_GREATERTHANOREQUAL', 162),
                   ('OP_MIN', 163),
                   ('OP_MAX', 164),
                   ('OP_WITHIN', 165),
                   ('OP_RIPEMD160', 166),
                   ('OP_SHA1', 167),
                   ('OP_SHA256', 168),
                   ('OP_HASH160', 169),
                   ('OP_HASH256', 170),
                   ('OP_CODESEPARATOR', 171),
                   ('OP_CHECKSIG', 172),
                   ('OP_CHECKSIGVERIFY', 173),
                   ('OP_CHECKMULTISIG', 174),
                   ('OP_CHECKMULTISIGVERIFY', 175),
                   ('OP_NOP1', 176),
                   ('OP_NOP2', 177),
                   ('OP_CHECKLOCKTIMEVERIFY', 177),
                   ('OP_NOP3', 178),
                   ('OP_CHECKSEQUENCEVERIFY', 178),
                   ('OP_NOP4', 179),
                   ('OP_NOP5', 180),
                   ('OP_NOP6', 181),
                   ('OP_NOP7', 182),
                   ('OP_NOP8', 183),
                   ('OP_NOP9', 184),
                   ('OP_NOP10', 185),
                   ('OP_NULLDATA', 252),
                   ('OP_PUBKEYHASH', 253),
                   ('OP_PUBKEY', 254),
                   ('OP_INVALIDOPCODE', 255), ]

    _opcode_to_int_dict = {op: num for op, num in opcodes}
    _int_to_opcode_dict = dict({num: op for op, num in opcodes})

    @classmethod
    def exists(cls, op):
        return op in cls._opcode_to_int_dict

    @classmethod
    def to_int(cls, op):
        try:
            return cls._opcode_to_int_dict[op]
        except KeyError:
            raise ValueError('Unknown op: {}'.format(op))

    @classmethod
    def from_int(cls, num):
        if num not in range(256):
            raise ValueError('Integer {} cannot be converted to opcode'.format(num))
        try:
            return cls._int_to_opcode_dict[num]
        except KeyError:
            return 'OP_UNKNOWN'

    @classmethod
    def to_hex(cls, op):
        return format(cls.to_int(op), '02x')

    @classmethod
    def from_hex(cls, hexnum):
        return cls.from_int(int(hexnum, 16))



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from binascii import unhexlify
from hashlib import sha256
from ecdsa import SigningKey, SECP256k1
from ecdsa.util import sigencode_der
from functools import partial
from abc import ABCMeta

from btcpy.lib.base58 import b58decode_check, b58encode_check
from btcpy.lib.types import HexSerializable
from btcpy.structs.address import P2pkhAddress, P2wpkhAddress


class WrongPubKeyFormat(Exception):
    pass


class Key(HexSerializable, metaclass=ABCMeta):
    pass


class InvalidWifLenght(Exception):
    '''Invalid WIF lenght.'''


class PrivateKey(Key):

    highest_s = 0x7fffffffffffffffffffffffffffffff5d576e7357a4501ddfe92f46681b20a0

    @staticmethod
    def from_wif(network, wif):

        if not 51 <= len(wif) <= 52:
            raise ValueError('Invalid wif length: {}'.format(len(wif)))

        decoded = b58decode_check(wif)
        prefix, *rest = decoded

        if prefix != network.wif_prefix:
            raise ValueError('Unknown private key prefix: {:02x}'.format(prefix))

        public_compressed = len(rest) == 33
        privk = rest[0:32]

        return PrivateKey(bytearray(privk), public_compressed)

    @staticmethod
    def unhexlify(hexa):
        return PrivateKey(bytearray(unhexlify(hexa)))

    def __init__(self, priv, public_compressed=True):
        self.key = priv
        self.public_compressed = public_compressed

    def to_wif(self, network):
        decoded = bytearray([network.wif_prefix]) + self.key
        if self.public_compressed:
            decoded.append(0x01)
        return b58encode_check(bytes(decoded))

    def pub(self, compressed=None):
        if compressed is None:
            compressed = self.public_compressed
        raw_pubkey = bytearray(SigningKey.from_string(self.key, curve=SECP256k1).get_verifying_key().to_string())
        uncompressed = PublicKey(bytearray([0x04]) + raw_pubkey)
        if compressed:
            return PublicKey(uncompressed.compressed)
        else:
            return uncompressed

    def serialize(self):
        return self.key

    def raw_sign(self, data, deterministic=True):
        sig_key = SigningKey.from_string(self.key, curve=SECP256k1)
        sig_func = partial(sig_key.sign_digest_deterministic, hashfunc=sha256) if deterministic else sig_key.sign_digest
        r, s, order = sig_func(data, sigencode=lambda *x: x)
        if s < 0x01:
            raise ValueError('Too low s value for signature: {}'.format(s))
        # ref: https://github.com/bitcoin/bips/blob/master/bip-0062.mediawiki#Low_S_values_in_signatures
        if s > PrivateKey.highest_s:
            s = order - s
        if s.to_bytes(32, 'big')[0] > 0x7f:
            s = int.from_bytes(b'\x00' + s.to_bytes(32, 'big'), 'big')
        if r.to_bytes(32, 'big')[0] > 0x7f:
            r = int.from_bytes(b'\x00' + r.to_bytes(32, 'big'), 'big')
        return r, s, order

    def sign(self, data, deterministic=True):
        return sigencode_der(*self.raw_sign(data, deterministic))

    def __eq__(self, other):
        return self.key == other.key

    def __str__(self):
        return self.hexlify()


class BasePublicKey(Key):

    p = 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f
    uncompressed_bytes = 64
    compressed_bytes = uncompressed_bytes // 2
    types = {0x02: 'even',
             0x03: 'odd',
             0x04: 'uncompressed',
             0x06: 'uncompressed',
             0x07: 'uncompressed'}

    uncompressed_headers = {header for header, type in types.items() if type == 'uncompressed'}
    compressed_headers = types.keys() - uncompressed_headers

    @staticmethod
    def from_point(point, compressed=True):
        result = PublicKey(bytearray([0x04]) + point.x().to_bytes(32, 'big') + point.y().to_bytes(32, 'big'))
        if compressed:
            return result.compress()
        return result

    @staticmethod
    def unhexlify(hexa):
        return PublicKey(bytearray(unhexlify(hexa)))

    @staticmethod
    def from_priv(priv):
        return priv.pub()

    def __str__(self):
        return self.hexlify()

    def __len__(self):
        return len(str(self)) // 2


class PublicKey(BasePublicKey):

    @classmethod
    def uncompress(cls, pubkey):
        header, *body = pubkey
        if header not in cls.compressed_headers:
            raise WrongPubKeyFormat('Pubkey header does not indicate compressed key: 0x{:02x}'.format(header))
        PublicKey.check(pubkey)
        parity = header - 2  # if 0x02 parity is 0, if 0x03 parity is 1
        x = int.from_bytes(body, 'big')
        alpha = (pow(x, 3, PublicKey.p) + 7) % PublicKey.p
        y = pow(alpha, (PublicKey.p + 1)//4, PublicKey.p)
        if y % 2 != parity:
            y = -y % PublicKey.p
        return bytearray([0x04]) + bytearray(body) + bytearray(y.to_bytes(PublicKey.compressed_bytes, 'big'))

    @classmethod
    def check(cls, pubkey):
        if type(pubkey) not in {bytes, bytearray}:
            raise ValueError('Unexpected data type for pubkey: {}'.format(type(pubkey)))

        try:
            header, *body = pubkey
        except ValueError:
            raise WrongPubKeyFormat('Got only one byte')

        if header in cls.uncompressed_headers:
            if len(body) != PublicKey.uncompressed_bytes:
                raise WrongPubKeyFormat('Unexpected length for uncompressed pubkey: {}'.format(len(body)))
        elif header in cls.compressed_headers:
            if len(body) != PublicKey.compressed_bytes:
                raise WrongPubKeyFormat('Unexpected length for compressed pubkey: {}'.format(len(body)))
        else:
            raise WrongPubKeyFormat('Unknown pubkey header: 0x{:02x}'.format(header))

    def __init__(self, pubkey):
        self.__class__.check(pubkey)
        self.type = PublicKey.types[pubkey[0]]
        if self.type == 'uncompressed':
            self.uncompressed = pubkey
            header = 0x03 if self.uncompressed[-1] % 2 else 0x02
            self.compressed = bytearray([header]) + self.uncompressed[1:-PublicKey.compressed_bytes]
        else:
            self.compressed = pubkey
            self.uncompressed = PublicKey.uncompress(pubkey)

    def hash(self):
        import hashlib
        original = self.uncompressed if self.type == 'uncompressed' else self.compressed
        sha = hashlib.sha256(original).digest()
        ripe = hashlib.new('ripemd160')
        ripe.update(sha)
        return bytearray(ripe.digest())

    def compress(self):
        if self.type != 'uncompressed':
            return self
        return PublicKey(self.compressed)

    def serialize(self):
        return self.uncompressed if self.type == 'uncompressed' else self.compressed

    def to_address(self, network):
        return P2pkhAddress(network, self.hash())

    def to_segwit_address(self, version):
        if self.type == 'uncompressed':
            pubk = PublicKey(self.compressed)
        else:
            pubk = self
        return P2wpkhAddress(pubk.hash(), version)

    def __eq__(self, other):
        return (self.type, self.compressed, self.uncompressed) == (other.type, other.compressed, other.uncompressed)

    def __lt__(self, other):
        return self.compressed < other.compressed




# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from abc import ABCMeta, abstractmethod

from binascii import hexlify
from functools import wraps


def cached(method):

    @wraps(method)
    def wrapper(self):
        if self.__class__._allow_caching:
            cache = '_{}_{}__cached'.format(self.__class__.__name__, method.__name__)
            if not hasattr(self, cache):
                # this is a workaround to allow caching of attributes also for immutable objects
                object.__setattr__(self, cache, method(self))
            return getattr(self, cache)
        return method(self)

    return wrapper


class Cacheable(metaclass=ABCMeta):

    _allow_caching = True


class Uncacheable(metaclass=ABCMeta):

    _allow_caching = False


class Immutable(Cacheable):

    def __setattr__(self, key, value):
        raise AttributeError('Trying to set attribute `{}` of immutable class `{}`'.format(key, self.__class__))

    def __delattr__(self, item):
        raise AttributeError('Trying to delete attribute `{}` of immutable class `{}`'.format(item, self.__class__))


class Mutable(Uncacheable):

    def __setattr__(self, key, value):
        object.__setattr__(self, key, value)

    def __delattr__(self, item):
        object.__delattr__(self, item)


class Jsonizable(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def from_json(cls, string):
        pass

    @abstractmethod
    def to_json(self):
        pass


class Serializable(metaclass=ABCMeta):

    @abstractmethod
    def serialize(self):
        raise NotImplemented


class Hexlifiable(metaclass=ABCMeta):

    @abstractmethod
    def hexlify(self):
        raise NotImplemented


class HexSerializable(Hexlifiable, Serializable, metaclass=ABCMeta):

    def hexlify(self):
        return hexlify(self.serialize()).decode()



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.

from abc import ABCMeta, abstractmethod
from .base58 import b58encode_check, b58decode_check

from .bech32 import decode, encode
from ..structs.address import Address, P2pkhAddress, P2shAddress, P2wpkhAddress, P2wshAddress


class CouldNotDecode(ValueError):
    pass


class CouldNotEncode(ValueError):
    pass


class Codec(metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def encode(address: Address) -> str:
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def decode(string: str, strict=None) -> Address:
        raise NotImplemented


class Base58Codec(Codec):

    hash_len = 20

    @staticmethod
    def encode(address):
        try:
            prefix = address.network.base58_raw_prefixes[address.get_type()]
        except KeyError:
            raise CouldNotEncode('Impossible to encode address type: {}, network: {}'.format(address.get_type(),
                                                                                             address.network))
        return b58encode_check(bytes(prefix + address.hash))

    @staticmethod
    def decode(network, string):

        try:
            addr_type = network.base58_prefixes[string[0]]
        except KeyError:
            raise CouldNotDecode('Impossible to decode address {}'.format(string))
        hashed_data = bytearray(b58decode_check(string))[1:]

        if len(hashed_data) != Base58Codec.hash_len:
            raise CouldNotDecode('Data of the wrong length: {}, expected {}'.format(len(hashed_data),
                                                                                    Base58Codec.hash_len))
        if addr_type == 'p2pkh':
            cls = P2pkhAddress
        elif addr_type == 'p2sh':
            cls = P2shAddress
        else:
            raise ValueError('Unknown address type: {}'.format(addr_type))

        return cls(network, hashed_data)


class Bech32Codec(Codec):

    lengths = {42: 'p2wpkh',
               62: 'p2wsh'}

    @staticmethod
    def encode(address):
        return encode(address.network.hrp, address.version, address.hash)

    @staticmethod
    def decode(network, string):

        if not string:
            raise CouldNotDecode('Impossible to decode empty string')

        lower = string[0].islower()
        for char in string:
            if not char.isdigit() and char.islower() != lower:
                raise CouldNotDecode('String {} mixes upper- and lower-case characters'.format(string))

        string = string.lower()
    
        if string[:2] != network.bech32_hrp:
            raise CouldNotDecode('Impossible to decode address {}'.format(string))

        try:
            addr_type = Bech32Codec.lengths[len(string)]
        except KeyError:
            raise CouldNotDecode('Impossible to decode address {}'.format(string))
        version, hashed_data = decode(string[:2], string)

        if not hashed_data:
            raise CouldNotDecode('Empty hash')

        if addr_type == 'p2wpkh':
            cls = P2wpkhAddress
        elif addr_type == 'p2wsh':
            cls = P2wshAddress
        else:
            raise ValueError('Unknown address type: {}'.format(addr_type))

        return cls(network, bytearray(hashed_data), version)



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.


from distutils.core import setup
from setuptools import find_packages

setup(name='peerassets-btcpy',
      version='0.5.2',
      packages=find_packages(),
      install_requires=['ecdsa==0.13'],
      extras_require={'develop': ['python-bitcoinlib==0.7.0']},
      description='A Python3 SegWit-compliant library which provides tools to handle Bitcoin data structures in a simple fashion.',
      author='chainside srl & PeerAssets development team',
      url='https://github.com/peerassets/btcpy',
      download_url='https://github.com/chainside/peerassets/archive/0.5.1.tar.gz',
      python_requires='>=3',
      keywords=['bitcoin', 'blockchain', 'peercoin', 'peerassets'])



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.


import subprocess
import errno
from time import sleep
import os
from shutil import copyfile, rmtree
from bitcoin.rpc import RawProxy, JSONRPCError


class NodesAlreadyConnected(Exception):
    pass


class RegtestRunning(Exception):
    pass


class SendCommandError(Exception):
    pass


class TxVerifyError(Exception):
    pass


class Manager(object):

    conf_path = './bitcoin.conf'
    base_port = 18300
    base_rpcport = 18400
    user = 'user'
    password = 'pwd'
    host = '127.0.0.1'

    def __init__(self, path='.'):
        self.nodes_number = 0
        self.nodes = {}
        self.regtest_path = path

    def generate_nodes(self, num):
        for n in range(num):
            self.gen_node(n)

    def gen_node(self, id_):
        dst_dir = '{}/node_{}'.format(self.regtest_path, id_)
        try:
            os.makedirs(dst_dir)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
        default_btc_conf = '{}/{}'.format(os.path.dirname(os.path.realpath(__file__)), Manager.conf_path)
        copyfile(default_btc_conf, '{}/{}'.format(dst_dir, 'bitcoin.conf'))
        ip_port = Manager.base_port + id_
        rcp_port = Manager.base_rpcport + id_
        rpc_user = Manager.user
        rpc_password = Manager.password
        with open('{}/{}'.format(dst_dir, 'bitcoin.conf'), 'a') as config_file:
            config_file.write("rpcuser={}\n".format(rpc_user))
            config_file.write("rpcpassword={}\n".format(rpc_password))
            config_file.write("port={}\n".format(ip_port))
            config_file.write("rpcport={}\n".format(rcp_port))
        self.nodes[id_] = None

    def start_nodes(self):
        for node in self.nodes:
            data_dir = os.path.abspath('{}/node_{}'.format(self.regtest_path, node))
            conf = os.path.abspath('{}/bitcoin.conf'.format(data_dir))
            cmd = ['bitcoind', '-conf={}'.format(conf), '-datadir={}'.format(data_dir),
                   '-maxtxfee=1000', '-debug=1', '-prematurewitness', '-zmqpubrawblock=tcp://127.0.0.1:28332']
            # cmd = ['/home/rael/bitcoin/src/bitcoind', '-conf={}'.format(conf), '-datadir={}'.format(data_dir),
            #        '-maxtxfee=100', '-debug=1', '-prematurewitness']
            subprocess.call(cmd)  # , stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

            self.nodes[node] = {'user': Manager.user,
                                'password': Manager.password,
                                'host': Manager.host,
                                'port': Manager.base_port + node,
                                'rpcport': Manager.base_rpcport + node}

        sleep(3)

        for n in self.nodes:
            for m in self.nodes:
                if m != n:
                    self.send_rpc_cmd(['addnode',
                                       '{}:{}'.format(Manager.host, self.nodes[m]['port']),
                                       'onetry'],
                                      n)

    def stop_nodes(self):
        self.send_rpc_cmd(['stop'], *[n for n in self.nodes])

    def teardown(self):
        from os.path import abspath
        self.stop_nodes()
        for n in self.nodes:
            rmtree('{}/node_{}'.format(abspath(self.regtest_path), n))

    def send_rpc_cmd(self, rpc_cmd, *nodes):

        # print('Sending rpc command: {}'.format(' '.join(rpc_cmd)))
        proxies = {}
        for n in nodes:
            proxies[n] = RawProxy('http://{user}:{password}@{host}:{rpcport}'.format(**self.nodes[n]))

        # try convert rpc params from strings to ints
        for idx, val in enumerate(rpc_cmd):
            try:
                rpc_cmd[idx] = int(val)
            except ValueError:
                if val in {'False', 'false'}:
                    rpc_cmd[idx] = False
                elif val in {'True', 'true'}:
                    rpc_cmd[idx] = True

        for node in nodes:
            retry = 1
            while True:
                try:
                    cmd, *args = rpc_cmd
                    # print("command:", rpc_cmd)
                    output = getattr(proxies[int(node)], cmd)(*args)
                    # print(output)
                    return output
                except JSONRPCError as e:
                    retry += 1
                    if e.error['code'] == -26:
                        raise TxVerifyError('Error during tx validation: {}'.format(e.error['message']))
                    if e.error['code'] != -28 or retry > 3:
                        raise SendCommandError("Error trying to send command to bitcoincli, "
                                               "error code: {}, message: {}".format(e.error['code'],
                                                                                    e.error['message']))
                    sleep(3 * (retry*2))



# Copyright (C) 2017 chainside srl
#
# This file is part of the btcpy package.
#
# It is subject to the license terms in the LICENSE.md file found in the top-level
# directory of this distribution.
#
# No part of btcpy, including this file, may be copied, modified,
# propagated, or distributed except according to the terms contained in the
# LICENSE.md file.


import sys
import copy
import unittest
from functools import partial
from argparse import ArgumentParser

from .regtest import Manager
from btcpy.structs.hd import ExtendedPrivateKey, ExtendedPublicKey
from btcpy.structs.transaction import Transaction, Sequence, TxOut, Locktime, TxIn, MutableTransaction, MutableTxIn
from btcpy.structs.sig import *
from btcpy.structs.script import *
from btcpy.setup import setup

setup('regtest')


keys = [('tpubDHVQPtNuLdRLj7FU348D5PcrkkPj5ibhN52cfjthEH9KTfwTaVmo'
         'dTn1Ekpge6PhUjW1noZ452xesirHgBKbzmY6hz4eoVXDwHcjczDT7zb',
         'tprv8koNFULfCFjfqeDg9QTcfyxkBisnvPQnnmRqPDrPp1LvdBggx6xD'
         'SyA94cjpAXS7ccGhsQ7w6q7Y9Ku31e1eTDztU49LVBz9B1sCDoeE6Jc'),
        ('tpubDEhdzhXujo86G6PXroPKQJSCJi8qbdQvrALhTNiExsGKfFHXtVbT'
         'E9tnLBCAP7nqQrqfUSVTCDuqv6RMHu8PDL5a8G43b5N2zKsF89nmLd6',
         'tprv8i1brHVfbRSRNdMjy9iiztn5jgcuSJE2GrjvArfwYbTvpm2mG6ms'
         '3fGvA5kdZ3qZ7KB26VehAudSCUURKT56Hej2pBgj26ZkNbdD1YMdTiD'),
        ('tpubDDqNYkcvEKbQJKtda5miuWCBWqX2Bd8qJWgSNbsiqfSHRzNpsjpX'
         'HAMiNYNHZw9FCnkuJpVAJjZkTeujhT4h293w6YMexGyAgNGRYWVtJ1D',
         'tprv8h9LQLag5wujQrrqgS78W6Y4wp162HwvjD5f65qRRPdtbW84FLzw'
         '6fjrCNeqEvsKqiDxLtzJ9oHUGVTL17KptjbDVqgJ2XvAs2LcvSWrTUh'),
        ('tpubDATAg7GX3FHknHSDzfVEgwo8U1aEWPNdkgRjs4dBF244x9xC6tRq'
         'UkM8ZMk8JNHnmQMqNG1evQBNKwt97G348FXaWjhT88UWbwqrBTpmwe3',
         'tprv8dm8XhEGtsc5tpQS71peHY91tz4JM4BjBNpxaYaspkFg7fhRUVcF'
         'JFjGPFHYGQLo21nmCrkjryoUFJKSKMUjKqpuRXAmMMhvMaP27UtqeLA'),
        ('tpubDHfgxSGzrfm7dWndtJ86CZPyoFNwkjB7dE2TNxp9BtwNcFq4oozW'
         '7bd5mCDzf1HBdXUofL1isdDA1LX7hAJmAx5Kp1grD85Q8UH5MCpuU51',
         'tprv8kyep2EkiJ5Sk3kqzeTVo9jsEDs1bPzD3vRg6Smqmd8ymmaJBRAu'
         'w71Db3Vuu8FThpRuLdoqhbvZZJ2zDNF7sNaJ6uB745RMESF5ezYNBAc'),
        ('tpubDExb5uGPg2pzURTo6aSQ1nf7eER9huQM96KmvbpY1AND3h12rfhV'
         'TEVeMQoT4Ljk22NQyTDt66eaJdjccT9JJZ8JGZQXq247b9SU7RjYCoP',
         'tprv8iGYwVE9Xf9KaxS1CvmocP115CuDYaDSZnize5nEatZpDCkGEGsu'
         'GjsnBGidvG6H3ZZGLqBvwjE2TF65vvMphKomStDGQWE5JBMsQA7UmZ7'),
        ('tpubDDLkGHUS4f3JrExQj1QN5LC4akrBW2VAhbLFMFC8mJgCpemZXifj'
         '4w2RQJHhZD4ET9VcxkcdZXxGfnUZv29EbsD7qD8rYJRrp9QXbenMXd7',
         'tprv8gei7sSBvHMdxmvcqMjmfvXx1jLFLhJG8HjU4j9qM2sozAWnuKr8'
         'tSQZEA57BdJkAqJKgFEb6S7StLBYpy4cxVDWZcVFSCcmfMXRdSUP5PR'),
        ('tpubDBD1DrcA8gpjpVdcbR5cFS9dHL9mUthxikEiVhmqXFqVHxr5tj6k'
         'Fa12HpJmfsiuMcWqRoq7FgHsKUSwwAKguWzUTwGak7SpR9suaUfuiD6',
         'tprv8eWy5SZuzK94w2bphmR1r2VWiJdqKZX49SdwDBjY6z36TUbKGLHA'
         '55PA7iUWTVKQjdQ95wz8Uy63pwuinehL72qsYyRFhM7i8huwRMxX9Jh'),
        ('tpubD9PSZUhp2M8CsZJMRXgg3MRp78bPM5L19fexkkiVxDaNcBUgUiPH'
         'TEFPg3rgbvaSxRjZGm793doYgRhV6kPpQNm7dinSniz8CrUaFc3JfwR',
         'tprv8chQR4fZsySXz6GZXt25dwmhY75TBk96aN4BUEgCXwmymhDurKZh'
         'GjdXVudZa6LhSSqo8vkPADFzssBtUXQKoRyK71yFaLEYG5fa1RGWKqY'),
        ('tpubD6NzVbkrYhZ4XjPybvVw1yfCMNhtBfcwfJaFycMouiGGhdrT5jyW'
         'b4NeivCte87ytt7f4qBaAQcWgTb3xS5oRrgnxByiQE7WjrkP1eZumon',
         'tprv8ZgxMBicQKsPeGNBiGqLca15nMBx2LS35zyUh6KWVSTss9bgTM9v'
         'QZknYnX2CdoDzkKmpRo8QZBvJceVBAAq7dSGK85j7MhCf2ZVToT36UX'),
        ('tpubDHFe2fdwqMEkzbTpx3GZjMSc1cfywRETR6dMx1Z7qoDHJgduZCr3'
         'Q6CkaJK4uDY9SpwonuDBTe8fjK9NCYmyzE5FQLvHCZEo5xhPYvwHgg8',
         'tprv8kZbtFbhgyZ678S34PbyKwnVSbA3n63Yqo2afVWpRXQtUCP8vp2T'
         'DbatQA23vW52PiDV52s8bnBMgBqQFJpGVdRi7RJkdxG9Yx32BPw6uKb'),
        ('tpubDEp72oRxJByNPNbhJp939pcwL67igJxHSzt6XQM7r5nF2uu3ZxPg'
         'S2hz3BsSVtvYAeweaJwTAJFgBzcRGvCgaTyVn31ut7CsvEnaaxp3Cwe',
         'tprv8i84tPPi9pHhVuZuRAUSkQxpm4bnWymNshHKEtJpRoyrCReGwZa6'
         'FY67s4cgwqxtbGNCWWWhcwcFWzdFDtpqeMpw96b56RtMfZXAFd9kKuh'),
        ('tpubDCZivWe1G5AxsNbupJnBPYUfBAy4eWbZs9BBCXV2rkbZ3N1Fi3e1'
         '9B1GcYeoT2UsngmqM2z1oPRwUUeEMuk1Y8Ku1bgrhNvcLiyg8ERkNoL',
         'tprv8fsgn6bm7hVHyua7vf7az8pYc9T8VBQfHqaPv1SjSUoACskV5epQ'
         'xgPQSQ1BjqXh4sjaZNZsxDvvjuwePEGHFBo1qbcZAvHixSZUfM4QQtr'),
        ('tpubDAtbo5g1qbYsLTD76KiRG4sNB48oD5eUTjgmPsnRMwuQSy4k5sag'
         'vJBH1LP5yGwDJHWhPQR61jqburWoee9kYuwkFJuTozddAVCFxQVB7wR',
         'tprv8eCZefdmhDsCSzBKCg3prfDFc2cs3kTZtS5z7Mk7wg71cUoyTUm6'
         'joZQqEHb87njVJxS8J1uNDXAB99mPPRixy4WpyZUbbYSij6mpbDpzKw'),
        ('tpubD8rXJ4n6Xbd5bP1zSJ1MKtkVEJapMQjiqJqdp3Tk8rAMsiWqt4i9'
         'HNt5omstF2pM6WL2Mu9R3VSxzFVGN5CMUG5PrkbGw9ukVjDyYXv4wTD',
         'tprv8cAV9ejrPDwQhuzCYeLkvV6NfH4tC5YpG1ErXXRSiaMy3EG5FftZ'
         '6tGDdbXmZ37Tw9MkZkMdMaHxL9LQGoPbyEWnGsEzesHegCfwZyQxQWH'),
        ('tpubD6NzVbkrYhZ4WM9KGkvWDfaQ8TbaigvHZnAcdyUH9BDihGTD3FkS'
         'Ed21ATMkqs4vTFY3LCgfeamCvRSnoQpsiaz8zmH5bYYiGMUUGKmnzyY',
         'tprv8ZgxMBicQKsPct7XP7FupFvHZS5eZMjNzUZqMTRyiuRKrnCSQrvr'
         '48Q8zLtc4w3vMCiZnmXXXVxxjRQFLZjgj7JHZ5n4WM2CULDEaqpCsVo')]

regtest = Manager()
regtest.generate_nodes(1)
regtest.start_nodes()
regtest.send_rpc_cmd(['generate', '500'], 0)

parser = ArgumentParser()
parser.add_argument('--dump', dest='dumpfile')
parser.add_argument('unittest_args', nargs='*')
cmdline_args = parser.parse_args()


def min_locktime(locktimes):
    m = 0
    res = Locktime(0)
    for lock in locktimes:
        if m is 0 or lock.n < m:
            m = lock.n
            res = lock
    return res


class Embedder(metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def get_script_cls():
        raise NotImplemented

    @staticmethod
    @abstractmethod
    def get_name():
        raise NotImplemented

    @staticmethod
    def get_args():
        return []

    def __init__(self, *args, scripts):
        self.instance = self.get_script_cls()(*(list(self.get_args()) + list(args) + list(scripts)))

    def post(self):
        pass


class P2shEmbedder(Embedder):
    @staticmethod
    def get_name():
        return 'p2sh'

    @staticmethod
    def get_script_cls():
        return P2shScript


class P2wshEmbedder(Embedder):
    @staticmethod
    def get_name():
        return 'p2wsh'

    @staticmethod
    def get_script_cls():
        return P2wshV0Script


class IfElseEmbedder(Embedder):
    @staticmethod
    def get_name():
        return 'ifelse'

    @staticmethod
    def get_script_cls():
        return IfElseScript


class TimelockEmbedder(Embedder):
    @staticmethod
    def get_name():
        return 'absolutetime'

    @staticmethod
    def get_script_cls():
        return TimelockScript

    @staticmethod
    def get_args():
        return Locktime(100),


class Relativetimelockembedder(Embedder):
    @staticmethod
    def get_name():
        return 'relativetime'

    @staticmethod
    def get_script_cls():
        return RelativeTimelockScript

    @staticmethod
    def get_args():
        return Sequence(3),

    def post(self):
        regtest.send_rpc_cmd(['generate', '3'])


class Hashlock160Embedder(Embedder):
    @staticmethod
    def get_name():
        return 'hash160'

    @staticmethod
    def get_script_cls():
        return Hashlock160Script


class Hashlock256Embedder(Embedder):
    @staticmethod
    def get_name():
        return 'hash256'

    @staticmethod
    def get_script_cls():
        return Hashlock256Script


class TestSpends(unittest.TestCase):

    @staticmethod
    def rand_bytes(n=500):
        import os
        return bytearray(os.urandom(n))

    @staticmethod
    def pairwise(iterable):
        from itertools import tee
        a, b = tee(iterable)
        next(b, None)
        return list(zip(a, b))

    def __init__(self, *args, **kwargs):
        global keys

        super().__init__(*args, **kwargs)

        pubs = [ExtendedPublicKey.decode(pair[0]).key for pair in keys]
        privs = [ExtendedPrivateKey.decode(pair[1]).key for pair in keys]
        all_embedders = {'p2sh', 'p2wsh', 'ifelse', 'absolutetime', 'relativetime', 'hash160', 'hash256'}

        self.scripts = [{'name': 'p2pkh',
                         'script': P2pkhScript(pubs[0]),
                         'solver': partial(P2pkhSolver, privs[0]),
                         'embeddable_by': all_embedders},
                        {'name': 'p2wpkh',
                         'script': P2wpkhV0Script(pubs[1]),
                         'solver': partial(P2wpkhV0Solver, privs[1]),
                         'embeddable_by': {'p2sh'}},
                        {'name': 'p2pk',
                         'script': P2pkScript(pubs[2]),
                         'solver': partial(P2pkSolver, privs[2]),
                         'embeddable_by': all_embedders},
                        {'name': 'multisig',
                         'script': MultisigScript(2, pubs[3], pubs[4], pubs[5], 3),
                         'solver': partial(MultisigSolver, privs[3], privs[4]),
                         'embeddable_by': all_embedders}, ]

        self.sighashed_scripts = []

        for script in self.scripts:
            for args in (('ALL', False), ('ALL', True), ('NONE', False), ('NONE', True), ('SINGLE', False),
                         ('SINGLE', True)):
                scriptcpy = copy.deepcopy(script)
                scriptcpy['sighash'] = Sighash(*args)
                try:
                    scriptcpy['solver'] = scriptcpy['solver'](sighash=Sighash(*args))
                except TypeError:
                    scriptcpy['solver'] = scriptcpy['solver'](sighashes=[Sighash(*args), Sighash(*args)])
                self.sighashed_scripts.append(scriptcpy)

        self.scripts = self.sighashed_scripts

        self.scripts = self.sighashed_scripts

        self.final = {'name': 'nulldata',
                      'script': NulldataScript(StackData.unhexlify('deadbeef')),
                      'solver': None,
                      'embeddable_by': {}}
        self.preimage_streams = [Stream(TestSpends.rand_bytes())]
        self.preimages = [pre.serialize() for pre in self.preimage_streams]
        self.hashes160 = [preimage.hash160() for preimage in self.preimage_streams]
        self.hashes256 = [preimage.hash256() for preimage in self.preimage_streams]

        self.all = [(s['script'],
                     (s['solver'], s['script']),
                     s['name']) for s in self.scripts]

        self.embedders = [TimelockEmbedder, Relativetimelockembedder,
                          Hashlock160Embedder, Hashlock256Embedder]

        self.double_embedders = [IfElseEmbedder]

        embedded = []

        for embedder in self.embedders:
            if embedder.get_name() != 'ifelse':
                for script in self.scripts:
                    if embedder.get_name() in script['embeddable_by']:
                        if embedder.get_name() == 'hash160':
                            for preimage, phash in zip(self.preimages, self.hashes160):
                                emb = embedder(phash, scripts=[script['script']])
                                embedded.append((emb.instance,
                                                 (HashlockSolver(preimage, script['solver']),
                                                  emb.instance),
                                                 '{}({})'.format(embedder.get_name(), script['name'])))
                        elif embedder.get_name() == 'hash256':
                            for preimage, phash in zip(self.preimages, self.hashes256):
                                emb = embedder(phash, scripts=[script['script']])
                                embedded.append((emb.instance,
                                                 (HashlockSolver(preimage, script['solver']),
                                                  emb.instance),
                                                 '{}({})'.format(embedder.get_name(), script['name'])))
                        else:
                            emb = embedder(scripts=[script['script']])
                            embedded.append((emb.instance,
                                             (TimelockSolver(script['solver']), emb.instance),
                                             '{}({})'.format(embedder.get_name(), script['name'])))

        self.all += [s for s in embedded]

        for embedder in self.double_embedders:
            included = [(x, y, t) for (x, y, t) in self.all if t != 'p2wpkh']
            for ((if_script, (if_solver, _), if_type), (else_script, (else_solver, _), else_type)) in TestSpends.pairwise(included):
                for branch in [Branch.IF, Branch.ELSE]:
                    inst = embedder(scripts=[if_script, else_script]).instance
                    # print(type(if_script), type(if_keys), type(if_spend), type(if_type))
                    # if 'hash' in else_type:
                    #     print(else_spend)
                    self.all.append((inst,
                                     (IfElseSolver(branch, if_solver if branch == Branch.IF else else_solver), inst),
                                     'ifelse({}, {})'.format(if_type, else_type)))

        for script, (solver, _), stype in [s for s in self.all]:
            if 'p2wpkh' not in stype:
                inst = P2wshEmbedder(scripts=[script]).instance
                self.all.append((inst,
                                 (P2wshV0Solver(script, solver), script),
                                 'p2wsh({})'.format(stype)))

        for script, (solver, prev), stype in [s for s in self.all]:
            inst = P2shEmbedder(scripts=[script]).instance
            self.all.append((inst,
                             (P2shSolver(script, solver), prev),
                             'p2sh({})'.format(stype)))

    def get_spending_data(self, solver, state=None):
        if state is None:
            state = {}
        # base case
        if isinstance(solver, (P2pkhSolver, P2wpkhV0Solver, P2pkSolver, MultisigSolver)):
            state['sig_hashes'] = [hexlify(sighash.as_byte()).decode() for sighash in solver.get_sighashes()]
            if isinstance(solver, P2pkSolver):
                state['priv_keys'] = [solver.privk.hexlify()]
            elif isinstance(solver, P2pkhSolver):
                state['priv_keys'] = [solver.privk.hexlify()]
            elif isinstance(solver, P2wpkhV0Solver):
                state['priv_keys'] = [solver.privk.hexlify()]
            else:
                assert isinstance(solver, MultisigSolver)
                state['priv_keys'] = [privk.hexlify() for privk in solver.privkeys]
            return state
        else:
            if isinstance(solver, HashlockSolver):
                return self.get_spending_data(solver.inner_solver, state)
            elif isinstance(solver, TimelockSolver):
                return self.get_spending_data(solver.inner_solver, state)
            elif isinstance(solver, IfElseSolver):
                try:
                    state['branches'].append(solver.branch.value)
                except KeyError:
                    state['branches'] = [solver.branch.value]
                return self.get_spending_data(solver.inner_solver, state)
            elif isinstance(solver, P2shSolver):
                return self.get_spending_data(solver.redeem_script_solver, state)
            elif isinstance(solver, P2wshV0Solver):
                return self.get_spending_data(solver.witness_script_solver, state)
            else:
                assert False

    def json_dump(self, unspent, spending, index, mutable_tx):
            dump = {'script_pubkey': {'hex': unspent['txout'].script_pubkey.hexlify(),
                                      'type': unspent['txout'].script_pubkey.type},
                    'spend_data': {'prev_amount': unspent['txout'].value}}
            prev_script = unspent['txout'].script_pubkey
            if isinstance(unspent['solver'], P2shSolver):
                dump['spend_data']['redeem_script'] = {'hex': unspent['solver'].redeem_script.hexlify(),
                                                       'type': unspent['solver'].redeem_script.type}
                prev_script = unspent['solver'].redeem_script
                if isinstance(unspent['solver'].redeem_script_solver, P2wshV0Solver):
                    dump['spend_data']['witness_script'] = {'hex': unspent['solver'].redeem_script_solver.witness_script.hexlify(),
                                                            'type': unspent['solver'].redeem_script_solver.witness_script.type}
                    prev_script = unspent['solver'].redeem_script_solver.witness_script
            elif isinstance(unspent['solver'], P2wshV0Solver):
                dump['spend_data']['witness_script'] = {'hex': unspent['solver'].witness_script.hexlify(),
                                                        'type': unspent['solver'].witness_script.type}
                prev_script = unspent['solver'].witness_script
            dump['script_sig'] = spending.script_sig.hexlify()
            if spending.witness is not None:
                dump['witness'] = spending.witness.hexlify()

            spend_data = dict(dump['spend_data'], **self.get_spending_data(unspent['solver']))
            dump['spend_data'] = spend_data
            dump['digests'] = []
            for sighash in unspent['solver'].get_sighashes():
                if unspent['solver'].solves_segwit():
                    dump['digests'].append(hexlify(mutable_tx.get_segwit_digest(index,
                                                                                prev_script=prev_script,
                                                                                prev_amount=unspent['txout'].value,
                                                                                sighash=sighash)).decode())
                else:
                    dump['digests'].append(hexlify(mutable_tx.get_digest(index, prev_script, sighash=sighash)).decode())
            return dump

    def test_all(self):
        global keys
        priv = ExtendedPrivateKey.decode(keys[0][1]).key
        pk = priv.pub()
        addr_string = str(pk.to_address())
        utxo = []

        for i in range(3):
            # create 3 tx to add to UTXO
            txid = regtest.send_rpc_cmd(['sendtoaddress', addr_string, '100'], 0)
            to_spend = Transaction.unhexlify(regtest.send_rpc_cmd(['getrawtransaction', txid, '0'], 0))
            txout = None
            for out in to_spend.outs:
                if str(out.script_pubkey.address()) == addr_string:
                    txout = out
                    break
            assert txout is not None

            utxo.append({'txid': txid,
                         'txout': txout,
                         'solver': P2pkhSolver(priv),
                         'next_seq': Sequence.max(),
                         'next_locktime': Locktime(0)})

        regtest.send_rpc_cmd(['generate', '100'], 0)

        generate = False
        next_locktime = Locktime(0)
        next_sequence = Sequence.max()

        i = 0
        while i < len(self.all) - 2:
            print('{:04d}\r'.format(i), end='', flush=True)
            ins = [MutableTxIn(unspent['txid'], unspent['txout'].n, ScriptSig.empty(), unspent['next_seq']) for unspent in utxo]
            outs = []
            prev_types = []

            for j, (unspent, script) in enumerate(zip(utxo, self.all[i:i+3])):
                outs.append(TxOut(unspent['txout'].value - 1000000, j, script[0]))
                prev_types.append(script[2])

            tx = MutableTransaction(2, ins, outs, min_locktime(unspent['next_locktime'] for unspent in utxo))
            mutable = copy.deepcopy(tx)
            tx = tx.spend([unspent['txout'] for unspent in utxo], [unspent['solver'] for unspent in utxo])

            # print('====================')
            # print('txid: {}'.format(tx.txid))
            # print()
            # print(tx)
            # print()
            # print('raw: {}'.format(tx.hexlify()))
            # print('prev_scripts, amounts, solvers:')

            print('TX: {}'.format(i))
            regtest.send_rpc_cmd(['sendrawtransaction', tx.hexlify()], 0)
            print('Mempool size: {}'.format(len(regtest.send_rpc_cmd(['getrawmempool'], 0))))

            if cmdline_args.dumpfile is not None:
                with open(cmdline_args.dumpfile, 'a') as out:
                    for j, unspent in enumerate(utxo):
                        json.dump(self.json_dump(unspent, tx.ins[j], j, copy.deepcopy(mutable).to_segwit()), out)
                        out.write('\n')

            utxo = []

            for j, (output, prev_type) in enumerate(zip(tx.outs, prev_types)):

                if 'time' in prev_type:
                    if 'absolute' in prev_type:
                        next_locktime = Locktime(100)
                        next_sequence = Sequence(0xfffffffe)
                    if 'relative' in prev_type:
                        next_sequence = Sequence(3)
                        generate = True
                else:
                    next_locktime = Locktime(0)
                    next_sequence = Sequence.max()

                utxo.append({'txid': tx.txid,
                             'txout': output,
                             'solver': self.all[i+j][1][0],  # solver
                             'next_seq': next_sequence,
                             'next_locktime': next_locktime})
            if generate:
                regtest.send_rpc_cmd(['generate', '4'], 0)
                generate = False

            if not i % 10:
                print('generating 2')
                regtest.send_rpc_cmd(['generate', '2'], 0)

            i += 1

        ins = [MutableTxIn(unspent['txid'],
                           unspent['txout'].n,
                           ScriptSig.empty(),
                           unspent['next_seq']) for unspent in utxo]

        tx = MutableTransaction(2,
                                ins,
                                [TxOut(sum(unspent['txout'].value for unspent in utxo) - 1000000, 0, self.final['script'])],
                                min_locktime(unspent['next_locktime'] for unspent in utxo))
        tx = tx.spend([unspent['txout'] for unspent in utxo], [unspent['solver'] for unspent in utxo])

        # print('====================')
        # print('txid: {}'.format(tx.txid))
        # print()
        # print(tx)
        # print()
        # print('raw: {}'.format(tx.hexlify()))
        # print('prev_scripts, amounts, solvers:')
        # for unspent in utxo:
        #     print(unspent['txout'].script_pubkey, unspent['txout'].value, unspent['solver'].__class__.__name__)
        regtest.send_rpc_cmd(['sendrawtransaction', tx.hexlify()], 0)

        regtest.teardown()


if __name__ == '__main__':
    sys.argv[1:] = cmdline_args.unittest_args
    unittest.main()
